/*
 * LED Panel to show dice rolling probabilities
 * Digital LED version - WS2812b "Neopixels" in a 16x16 sheet on pin D6
 * 6/2021 Carl F Sutter
 * Processor is an ESP8266 (Wemos D1 Mini)
 * Push button connecting pin 0 (D3) to ground
 */

#include <Adafruit_NeoPixel.h>
#define PIN        D6   // I/O pin connected to the LED data line
#define NUMPIXELS 256   // number of LEDs in the array/strand
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

#define DICE_MAX 6                 // maximum dice roll - will need to add one with the random function since it goes from 0 to n-1
#define TWO_DICE_MAX DICE_MAX * 2  // two dice range actually 11 values 2 (two ones) - 12 (two sixes)
#define long_press_delay 200       // the ms delay between button presses when you hold the button down
#define text_scroll_delay 200      // the ms delay between srolling text steps
#define PIN_BUTTON1 0              // D3 - can't seem to use the ESP8266 constants here

#define MODE_SINGLE   0       // do a single dice roll per step
#define MODE_MULTIPLE 1       // run many rolls per step
int tally_mode = MODE_SINGLE; // start out in single mode - one dice roll per push button press
#define MAX_TALLY 15          // maximum tally (number of times for a specific dice roll) to stop animation - equal to the numbver of LEDs in a column
#define ROLLS_MULTIPLE 1000   // number of rolls per click in multiple mode

int tallies_single[TWO_DICE_MAX];              // stores the tallies of each dice roll in single roll mode
unsigned int tallies_multiple[TWO_DICE_MAX];   // stores the tallies of each 1000 dice rolls in multiple roll mode
bool rolls_done = false;                       // set to false to run the dice roll simulations and charting - gets set to true when the max tally is reached - set to true to not run the dice roll part
uint32_t draw_color = pixels.Color(16, 0, 0);  // used to globally set the color for drawing on the screen

int multiple_step_num = 0;  // count the number of times the button was pressed in multiple mode for UI feedback

String strRoll1 = "1 ROLL PER PRESS";         // scrolling text for single mode
String strRoll1000 = "1000 ROLLS PER PRESS";  // scrolling text for multiple mode
String strCurrent = strRoll1;


//********************************************************************************
// setup - runs once on startup
//********************************************************************************
void setup() {
  //Serial.begin(9600);

  // init all the roll tally counts to zero
  clear_tallies();

  // help ensure a different result each time - true random numbers are a whole thing
  randomSeed(analogRead(0));

  // push button will be active low - use internal pullup
  pinMode(PIN_BUTTON1, INPUT_PULLUP);

  pixels.begin(); // INITIALIZE NeoPixel strip object (REQUIRED)
  startup_screen();
} // setup


//********************************************************************************
// loop - runs repeatedly
//********************************************************************************
void loop() {
  // if the previous run is complete and the button is pressed, reset for the next run
  if (rolls_done && (digitalRead(PIN_BUTTON1) == LOW)) {
    clear_tallies();
    rolls_done = false;
    startup_screen();
    delay(500);
    return;
  }

  // if we are in a run and the button was pressed, roll the dice
  if (!rolls_done && (digitalRead(PIN_BUTTON1) == LOW)) {
    pixels.clear(); // Set all pixel colors to 'off'

    // in single mode, roll the dice once
    if (tally_mode == MODE_SINGLE) {
      // want two dice, each 1-6, so final val would be 2-12
      // roll each die and display
      draw_color = pixels.Color(0, 8, 0);
      int roll1 = random(DICE_MAX) + 1;
      draw_dice(roll1, 0, 15);
      int roll2 = random(DICE_MAX) + 1;
      draw_dice(roll2, 0, 11);

      // compute the sum, display it, and add to the tallies
      draw_color = pixels.Color(0, 0, 8);
      int sum = roll1 + roll2;
      if (sum > 9) draw_digit(1, 13, 15);
      draw_digit(sum % 10, 13, 9);
      int idx = sum - 2;  // array index is 0-12, so a roll of 2 would be in the 0th element etc.
      tallies_single[idx]++;
      // stop if one of the tallies hit 16, the max value
      if (tallies_single[idx] == MAX_TALLY) rolls_done = true;

      // display the tallies
      draw_color = pixels.Color(16, 0, 0);
      for (int i=0; i<TWO_DICE_MAX-1; i++) set_led(i+2, tallies_single[i]);  // i+1 is to shift it right since we have 2-12 on a 16 pixel wide display
      pixels.show();   // Send the updated pixel colors to the hardware.
    } // MODE_SINGLE

    // in multiple mode, roll the dice 1000 times per click
    if (tally_mode == MODE_MULTIPLE) {
      for (int i=0; i<ROLLS_MULTIPLE; i++) {
        int roll1 = random(DICE_MAX) + 1;
        int roll2 = random(DICE_MAX) + 1;
        int sum = roll1 + roll2;
        int idx = sum - 2;  // array index is 0-12, so a roll of 2 would be in the 0th element etc.
        tallies_multiple[idx]++; // array index is 0-12, so a roll of 2 would be in the 0th element etc.
        if (tallies_multiple[idx]/ROLLS_MULTIPLE >= MAX_TALLY) rolls_done = true;
      }
      // after the rolls, update the chart
      // draw a green dot to let the user know the button press happened since tghe chart does not always change
      draw_color = pixels.Color(0, 0, 16);
      set_led(15, multiple_step_num++ % 16);
      // update the tally chart
      draw_color = pixels.Color(16, 0, 0);
      for (int i=0; i<TWO_DICE_MAX-1; i++) set_led(i+2, tallies_multiple[i]/ROLLS_MULTIPLE);
      pixels.show();   // Send the updated pixel colors to the hardware.
    }
    
    // pause between steps - this slows it down when the button is held down
    delay(long_press_delay);

    // if the run is complete (one tally hit the top), draw 10 dots on the left side, then remove them one by one as a count down
    if (rolls_done) {
      multiple_step_num = 0;
      draw_color = pixels.Color(0, 8, 8);
      for (int i=0; i<9; i++) set_led(0, i);
      pixels.show();
            
      draw_color = pixels.Color(0, 0, 0);
      for (int i=0; i<9; i++) {
        set_led(0, 8-i);
        pixels.show();
        delay(500);
      }

      // switch modes when we complete one mode
      if (tally_mode == MODE_SINGLE) {
        tally_mode = MODE_MULTIPLE;
        strCurrent = strRoll1000;
      } else {
        tally_mode = MODE_SINGLE;      
        strCurrent = strRoll1;
      }
    } // rolls_done
  } // (!rolls_done && (digitalRead(PIN_BUTTON1) == LOW)
  
} // loop


//********************************************************************************
// draw_digit - draw a 3x5 dot number at the given x & y coordinate - upper left corner, origin is lower left
//********************************************************************************
void draw_digit(int digit, int col, int row) {
  switch (digit) {
    case 0:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 1:
      set_led(col+1, row);
      set_led(col, row-1);
      set_led(col+1, row-1);
      set_led(col+1, row-2);
      set_led(col+1, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 2:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row-1);
      set_led(col+1, row-2);
      set_led(col, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 3:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col+2, row-1);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 4:
      set_led(col, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col+2, row-3);
      set_led(col+2, row-4);
      break;
    case 5:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 6:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 7:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col+2, row-1);
      set_led(col+1, row-2);
      set_led(col+1, row-3);
      set_led(col+1, row-4);
      break;
    case 8:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 9:
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
  }
} // draw_digit


//********************************************************************************
// draw_char - draw a 3x5 dot character at the given x & y coordinate - upper left corner, origin is lower left
//********************************************************************************
void draw_char(char cdraw, int col, int row) {
  // only doing necessary characters
  switch (cdraw) {
    case '0':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case '1':
      set_led(col+1, row);
      set_led(col, row-1);
      set_led(col+1, row-1);
      set_led(col+1, row-2);
      set_led(col+1, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 'E':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 'L':
      set_led(col, row);
      set_led(col, row-1);
      set_led(col, row-2);
      set_led(col, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 'O':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 'P':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col, row-4);
      break;
    case 'R':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col, row-3);
      set_led(col+1, row-3);
      set_led(col, row-4);
      //set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
    case 'S':
      set_led(col, row);
      set_led(col+1, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col, row-2);
      set_led(col+1, row-2);
      set_led(col+2, row-2);
      set_led(col+2, row-3);
      set_led(col, row-4);
      set_led(col+1, row-4);
      set_led(col+2, row-4);
      break;
  }
} // draw_char


//********************************************************************************
// draw_dice - draw a dice face at the given x & y coordinate - upper left corner, origin is lower left
//********************************************************************************
void draw_dice(int digit, int col, int row) {

  switch (digit) {
    case 1:
      set_led(col+1, row-1);
      break;
    case 2:
      set_led(col+2, row);
      set_led(col, row-2);
      break;
    case 3:
      set_led(col+2, row);
      set_led(col+1, row-1);
      set_led(col, row-2);
      break;
    case 4:
      set_led(col, row);
      set_led(col+2, row);
      set_led(col, row-2);
      set_led(col+2, row-2);
      break;
    case 5:
      set_led(col, row);
      set_led(col+2, row);
      set_led(col+1, row-1);
      set_led(col, row-2);
      set_led(col+2, row-2);
      break;
    case 6:
      set_led(col, row);
      set_led(col+2, row);
      set_led(col, row-1);
      set_led(col+2, row-1);
      set_led(col, row-2);
      set_led(col+2, row-2);
      break;
  }
} // draw_dice


//********************************************************************************
// startup_screen - draw a startup screen
//********************************************************************************
void startup_screen() {
  //pixels.clear(); // Set all pixel colors to 'off'
  //draw_color = pixels.Color(0, 8, 8);
  //for (int i=0; i<16; i++) set_led(i, 0);
  
  // wait for the pushbutton to be released
  while (digitalRead(PIN_BUTTON1) == LOW) delay(1);
  
  pixels.clear(); // Set all pixel colors to 'off'
  draw_color = pixels.Color(2, 8, 0);
  bool bStop = false;
  while (!bStop) {
    // stepping from the right of the screen to all the way left
    for (int i=15; i>-(int)strCurrent.length()*4; i--) {
      pixels.clear(); // Set all pixel colors to 'off'
      for (int j=0; j<strCurrent.length(); j++) {
        draw_char(strCurrent.charAt(j), i+(j*4), 10);
        // stop if the button was pressed
        if (bStop || (digitalRead(PIN_BUTTON1) == LOW)) {
          bStop = true;
          break;
        }
      }
      pixels.show();   // Send the updated pixel colors to the hardware.
      // stop if the button was pressed
      if (bStop || (digitalRead(PIN_BUTTON1) == LOW)) {
        bStop = true;
        break;
      }
      delay(text_scroll_delay);
    }
  } // display while loop

  // button was pressed to stop, scrolling text is done - display a static screen
  pixels.clear(); // Set all pixel colors to 'off'
  draw_color = pixels.Color(8, 0, 0);
  for (int i=2; i<13; i++) set_led(i, 0);
  pixels.show();

  // wait for the button to continue
  while (digitalRead(PIN_BUTTON1) == LOW) delay(1);

} // startup_screen



//********************************************************************************
// set_led - turn on the LED at the given x & y coordinate
//********************************************************************************
void set_led(int col, int row) {
  int led_num = row * 16;

  // stop if the pixel is out of range - scrolling text will do this
  if ((col < 0) || (col > 15) || (row < 0) || (row > 15)) return;
  
  // the pixels go in a serpentine direction, so we have to reverse the x coord of odd rows
  if ((row % 2) == 0){
    led_num += col;
  } else {
    led_num += 15 - col;
  }
  pixels.setPixelColor(led_num, draw_color);
} // set_led


//********************************************************************************
// clear_tallies - init all the roll tally counts to zero
//********************************************************************************
void clear_tallies() {
  for (int i=0; i<TWO_DICE_MAX; i++) {
    tallies_single[i] = 0;
    tallies_multiple[i] = 0;
  }
} // clear_tallies
